
/*	Copyright (c) 1982 Dyadic Systems Limited	*/

/*
	l p a p l  -  a print spooler backend for dyalog apl

	This is a backend for dqueuer(M).   It does not strip control codes.  It
	is intended  for printing output  that includes  SO and SI to switch the
	APL character  set of  the printer.   The  control character  SI is sent
	first to force the printer out of APL.   A form-feed preceeds the banner
	and each copy requested.  As described in BACKEND(M) the banner includes
	the first word found in /etc/sitefile, if this is accessible.
*/

#include <stdio.h>
#include <spool.h>
#include <time.h>

#define	FF	'\014'	/* form feed			*/
#define	SI	'\017'	/* go non-apl			*/
#define PW	132	/* page width for banner	*/

	/* indexes into argv of corresponding arg, or zero */
int	nofil,	ncops,	title,	from,	dest,	file,	spool;
FILE	*outfile;

main(argc,argv)
int	argc;
char	*argv[];
{
	int	ncopies;
	int 	c;
	FILE	*infile;
  
	outfile = fdopen(SFFD,"a");
	decodeargs(argc,argv);
	banner(argv);

	/* default 1 copy, but set to zero if no file to be printed */
	ncopies = (0!=nofil)?0 : (0==ncops)?1 : atoi (argv[1+ncops]);
	if (ncopies) infile = fdopen(RFFD,"r");
	while(ncopies--)
		{putc(FF,outfile);
		while (EOF != (c=getc(infile))) putc(c,outfile);
		if(ncopies) rewind(infile);
		}
	fclose(outfile);
}

/*
	print banner page
*/
banner(argv)
char	*argv[];
{
	char	*printtime;
	long	time(),thistime;
	char 	*ctime();
	char	buf[2+PW];
	char	flash[PW];

	thistime=time(0);
	printtime=ctime(&thistime);
	putc(FF,outfile);
	putc(SI,outfile);

	setflash(flash);
	flasher (4,flash);
	newlines (14);

	/* invoke banner from library to do big letters */
	fflush (outfile);
	strcpy (buf,"banner '");
	strcat (buf,argv[1+title]);
	strcat (buf,"' >&");
	sprintf(buf,"%s%d",buf,SFFD);
	system (buf);
	fflush (outfile);

	/* print file details */
	newlines(11);
	flasher (1,"***");
	prtbuf(buf," "," ");
	prtbuf(buf,"Print spooler"," ");
	prtbuf(buf," "," ");
	prtbuf(buf,"FROM:",argv[1+from]);
	prtbuf(buf,"TO:  ",argv[1+dest]);
	prtbuf(buf,"FILE:",argv[1+file]);
	prtbuf(buf,"TIME SPOOLED:",argv[1+spool]);
	prtbuf(buf,"TIME PRINTED:",printtime);
	prtbuf(buf," "," ");
	flasher (1,"***");
	newlines(13);
	flasher (4,flash);
}

/*
	set globals to index corresponding arg
*/
decodeargs(argc,argv)
int	argc;
char	*argv[];
{
	while(--argc)
		if('-'==*argv[argc])
			switch(*(1+argv[argc]))
				{case 'F': from  = argc; break;
				case 's' : spool = argc; break;
				case 'f' : file  = argc; break;
				case 'B' : nofil = argc; break;
				case 'c' : ncops = argc; break;
				case 'd' : dest  = argc; break;
				case 't' : title = argc; break;
				default  : break;
				}
	if(dest ==0) dest=from;	/*dest defaults to from */
	if(title==0)title=dest;	/*then title defaults to dest */
}

/*
	print n lines of string s
*/
flasher(n,s)
int	n;
char	*s;
{
	int	i, l, b, d;

	l=strlen(s);
	b=0;
	while(n--)
		{for(d=0,i=b ;d<PW ; d++)
			{putc(s[i++],outfile);
			if(i >= l) i=0;
			}
		putc('\n',outfile);
		if( ++b >= l ) b=0;
		}
}

/*
	print n new lines
*/
newlines(n)
int	n;
{
	while(n--)putc('\n',outfile);
}

/*
	set flash to first word in sitefile or default
*/
setflash(flash)
char	*flash;
{
	FILE	*sitefile;
	int 	c;

	strcpy(flash,"DYALOG ");
	if(NULL==(sitefile=fopen("/etc/sitefile","r")))return;
	while((c=getc(sitefile))!=EOF && c!=' ' && c!='\t' && c!='\n')
		*flash++ = c;
	*flash++ = ' ';
	*flash++ = '\0';
	fclose(sitefile);
}

/*
	insert strings into blanked out buf and print
*/
prtbuf(buf,str1,str2)
char	*buf,	*str1,	*str2;
{
	char	*start,	*last;

	start = buf;
	last = buf + PW-1;
	*buf++ = '*', *buf++ = ' ', *buf++ = ' ';
	while(*str1) *buf++ = *str1++;
	*buf++ = ' ';
	while(*str2 && *str2 != '\n') *buf++ = *str2++;
	while(buf < last) *buf++ = ' ';
	*buf++ = '*', *buf++ = '\n' , *buf++ = '\0';
	fputs(start, outfile);
}

/*	Copyright (c) 1982 Dyadic Systems Limited	*/


