
/*	Copyright (c) 1982 Dyadic Systems Limited	*/

#include	"apl.h"
#include	"support.h"
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<errno.h>

#define		PIPESIZE	4096

jmp_buf		errbuff;
static jmp_buf	env15;		/* buf for handshake */
static int	inerror;	/* flag for handshake control */
unsigned	parentid;


/*    s t a r t u p    */
/*	startup protocol as in "apl.h"					*/
startup(n_fns)
{
	toapl(getpid());
	parentid = fromapl();
	toapl(n_fns);
	set15();
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGHUP, SIG_DFL);
}


/*	d e f i n e	*/
define(apfn, syntax, fncode)
char *apfn;			/* function name to appear in workspace	*/
int syntax, fncode;		/* 'syntax'  as defined in "apl.h".	*/
{
	register int length;

	toapl(length = strlen(apfn));
	write(RSLTPIPE, apfn, length);
	toapl(syntax);
	toapl(fncode);
}


/*	g e t a r r a y   -   p u t a r r a y	*/

ARRAY *getarray()		/* get(possibly nested) array from apl	*/
{
	register unsigned rank, *shape;
	register ARRAY *array, **ptr;
	unsigned nwords;	/* number of words remaining to read	*/
	extern int read();

	if((nwords = fromapl()) == 0)	/* read word count		*/
		return(0);		/* left arg omitted		*/

	array = alloc(--nwords, ARRAY);	/* adjust word count &		*/
					/* allocate space for array	*/
	shape = &(array->shape[0]);	/* take pointer to shape	*/
	array->length = nwords;		/* fill in array length		*/

	fields(array) = fromapl();	/* read type/rank/eltype	*/
	nwords -= 2;			/* decr words to be read	*/

	if(array->type == SIMPLE)
		pipewds(read, ARGSPIPE, shape, nwords);	/* simple array */
	else
	{
		rank = array->rank;
		pipewds(read, ARGSPIPE, shape, rank);	/* read shape	*/
		nwords -= rank;			/* ... and decr count	*/

		/*	now recursively read in sub-arrays and
			install pointers in parent array.
		*/
		for(ptr = data(array, ARRAY *); nwords--;)
			*ptr++ = getarray();
	}
	return array;			/* return pointer to array	*/
}


putarray(array)		/* pass (possibly nested) array to apl		*/
ARRAY *array;		/* arg is pointer to array to be transferred	*/
{
	register ARRAY **ptr;
	register unsigned rank, nwords;
	extern int write();

	if(array == 0)
		toapl(0);			/* null result		*/
	else
	{
		toapl(nwords = array->length+1);	/* pipe length word */
		if(array->type == SIMPLE)
			pipewds(write, RSLTPIPE, &fields(array), nwords-2);
		else
		{
			rank = array->rank;
			/* pipe and shape */
			pipewds(write, RSLTPIPE, &fields(array), rank+1);
			nwords -= rank+1+2;	/* words remaining	*/
			/* recursively put each subarray		*/
			for(ptr = data(array, ARRAY *); nwords--;)
				putarray(*ptr++);
		}
	}
}


/*  t h r o w a r r a y   -   f r e e a r r a y  */

ARRAY *throwarray(thing)			/* send the array to	*/
ARRAY *thing;					/* APL & free the space	*/
{
	putarray(thing);
	return(freearray(thing));
}


ARRAY *freearray(thing)				/* return an ARRAY to	*/
ARRAY *thing;					/* the free space pool	*/
{
	unsigned n, n_elts = 1;
	ARRAY **ptr;

	if(thing == 0)				/* don't try to free	*/
		return(0);			/* if arg is null ptr	*/

	if(thing->type != SIMPLE)		/* recursively free	*/
	{					/* subarrays first	*/
		for(n = thing->rank; n--;)
			n_elts *= thing->shape[n];

		ptr = &data(thing, ARRAY *)[n_elts];
		while(n_elts--)			/* free subarrays in	*/
			freearray(*--ptr);	/* reverse order	*/
	}

	free((char *)thing);			/* free this array	*/
	return(0);
}


/*	m k a r r a y	*/

static unsigned char SHIFT[] =	/* to convert no. of elements to no. of bits */
{	3, 	/* CHAR */
	0, 	/* BOOL */
	3, 	/* SINT */
	4, 	/* INTG */
	5, 	/* LONG */
	6, 	/* DOUB */
	WDBITS	/* PNTR */
};

ARRAY *mkarray(eltype, rank, shape)		/* make an array in memory */
unsigned eltype;				/* element type */
unsigned rank;					/* array rank */
unsigned *shape;				/* pointer to shape vector */
{
	register unsigned bound = 1, nwds, u, *ptr;
	register ARRAY *array;

	for(ptr = shape, u = rank; u--;)	/* calc no of elements as */
		bound *= *ptr++;		/* product of shape words */

	nwds =					/* calc no of words needed */
		2				/* length and */
		+ rank +			/* shape vector		*/
		(				/* array elements	*/
			(
				((long)bound << SHIFT[eltype])	/* no of bits */
				+(1 << WDBITS)-1		/* rounded up */
			) >> WDBITS		/* converted to words	*/
		) +
		(eltype == PNTR && bound == 0);	/* prototype if null nested */

	array = alloc(nwds, ARRAY);		/* allocate space for array */
						/* and fill in type ...	*/
	array->type = (eltype == PNTR ? NESTED : SIMPLE);
	array->length = nwds;			/* length		*/
	array->eltype = eltype;			/* element type		*/
	array->rank = rank;			/* rank			*/

	for(ptr = &array->shape[0]; rank--;)
		*ptr++ = *shape++;		/* install shape	*/

	return array;
}


/*	u t i l i t i e s	*/

error(errnumber)		/* signal error to apl		      */
int errnumber;
{
	int (*savequit)();	/* save current strong interrupt */
	int (*saveint)();	/* save current weak interrupt	 */

	savequit=signal(SIGQUIT,SIG_IGN); /* dare not interrupt handshake */
	saveint=signal(SIGINT,SIG_IGN);
	
	inerror=1;		/* so take15() knows who instigated */
	if(!setjmp(env15))
		{
		 kill(parentid,15);	/* stop apl		*/
		 pause();	/* wait for apl to clean the result pipe*/
		}
	
	if(!setjmp(env15))
		{
		 toapl(errnumber);	/* write one word error code */
		 pause();		/* wait for signal from apl  */
		}
	gobble(ARGSPIPE);		/* clean input pipe	*/
	kill(parentid,15);		/* tell apl you have finished */

	signal(SIGQUIT,savequit);	/* restore interrupts	*/
	signal(SIGINT,saveint);

	inerror=0;
	longjmp(errbuff,errnumber);	/* go back for next object */
}

static gobble(pipe)		/* consume any bytes queued on pipe	*/
unsigned pipe;
{
	struct stat buf;
	char c;
	register unsigned count;

	while(1)
	{
	 	if(fstat(pipe, &buf) == -1)
			return;			/* can't stat pipe	*/
	 	if((count = buf.st_size) == 0)
			return;			/* finished		*/
		while(count--)
			read(pipe, &c, 1);	/* throw away bytes	*/
	}
}

static take15()
{
	set15();
	if(inerror)longjmp(env15, 1);	/* return to error */
	gobble(ARGSPIPE);	/* apl issued an error */
	kill(parentid,15);	/* tell apl you have finished */
	longjmp(errbuff,1);	/* go get next input	*/
}

set15()
{
	signal(15, take15);
}

static pipewds(iofn, pipe, buff, size)		/* pipe words to or from ap */
int (*iofn)();					/* read or write	*/
unsigned pipe;					/* pipe number		*/
unsigned *buff;					/* word ptr		*/
unsigned size;					/* size in wds		*/
{
	register unsigned chunk, pipesize;
	register unsigned char *cbuf;
	register int nbytes;
	extern int errno;

	size *= sizeof(int);			/* convert to bytes	*/
	cbuf = (unsigned char *)buff;
	pipesize = PIPESIZE - sizeof(int);

	while(chunk = size < pipesize ? size : pipesize)
	{
		/* transfer a chunk at a time */

		nbytes = (*iofn)(pipe, cbuf, chunk);
		if(nbytes == chunk){		/* success ! */
			size -= chunk;
			cbuf += chunk;
			pipesize = PIPESIZE;
		}
		else if (nbytes == 0)
		{
			kill(getpid(), SIGHUP);
			exit(0);		/* EOF => apl finished */
		}
		else if (nbytes==-1 && errno==EINTR)
			continue;	/* io interrupted try again */
		else
			exit(errno);	/* give up */
	}
}


toapl(n)
unsigned n;
{
	pipewds(write, RSLTPIPE, &n, 1);
}


unsigned fromapl()
{
	unsigned rslt;

	pipewds(read, ARGSPIPE, &rslt, 1);
	return(rslt);
}


char *_zalloc(nbytes)			/* same as C library 'malloc'	*/
unsigned nbytes;			/* but with error check		*/
{
	char *p, *malloc();

	if((p = malloc(nbytes)) == 0)
		error(LIMIT);

	return(p);
}

/*	Copyright (c) 1982 Dyadic Systems Limited	*/


