
/*	Copyright (c) 1982 Dyadic Systems Limited	*/

/*	Unix file functions	*/

#include "apl.h"

jmp_buf errbuff;

main()
{
	ARRAY	*rslt, *larg, *rarg, *getarray(), *freearray();
	ARRAY	*xclose(),
		*xcreat(),
		*xlseek(),
		*xopen(),
		*xread(),
		*xwrite(),
		*xavtounix(),
		*xunixtoav();
	ARRAY	*(*fromapl())(),
		*(*function)();

	startup(8);		/* signal number of functions to apl */

	define("close",	RSLT|OPTL , xclose);
	define("creat",	RSLT|OPTL , xcreat);
	define("lseek",	RSLT|OPTL , xlseek);
	define("open",	RSLT|OPTL , xopen);
	define("read",	RSLT|OPTL , xread);
	define("write",	RSLT|OPTL , xwrite);
	define("avtounix", RSLT|OPTL , xavtounix);
	define("unixtoav", RSLT|OPTL , xunixtoav);

	larg = rarg = 0;

	setjmp(errbuff);			/* come here after error */

	while (1) {
		larg = freearray(larg);		/* free up heap space	*/
		rarg = freearray(rarg);

		function = fromapl();		/* get desired function	*/
		larg = getarray();		/* get arguments	*/
		rarg = getarray();
		
		rslt = (*function)(larg,rarg);	/* run the function	*/
		putarray(rslt);			/* return result	*/
	}
}

ARRAY	*mkintg(n)
int	n;
{
	static ARRAY	temp;

	temp.length = 3;
	temp.type = SIMPLE;
	temp.rank = 0;
	temp.eltype = DINT;
	temp.shape[0] = (unsigned)n;
	return &temp;
}

/*	g e t f i l e n a m e

	return pointer to file name string - or error
*/

unsigned char	*getfilename(array)
ARRAY		*array;
{
	static unsigned char	filename[256];
	unsigned char		*from, *to, avtounix();
	unsigned		shape;

	if(CHAR != array->eltype) error(DOMAIN);
	if(1 != array->rank) error(RANK);
	shape=array->shape[0];
	if(0==shape || sizeof(filename)<=shape) error(LENGTH);

	to=filename;
	from= (unsigned char*) (1+array->shape);
	while(shape--) *(to++) = avtounix(*(from++));
	*to = '\0';

	return filename;
}

/*	g e t i n t

	return the value of a simple integral scalar or singleton vector
	- or domain error
*/

int	getint(array)
ARRAY	*array;
{
	int	(*makeint)(), (*initint())();

	if(array->rank)
		if(array->rank > 1) error(RANK);
		else if(array->shape[0] != 1) error(LENGTH);

	makeint = initint(array);
	return (*makeint)(array->shape+array->rank, 0);
}

/*	c l o s e

	z <- close fildes
*/

ARRAY	*xclose(larg,rarg)
ARRAY	*larg, *rarg;
{
	if(larg) error(SYNTAX);
	return mkintg(close(getint(rarg)));
}

/*	c r e a t e

	fildes <- {mode} create name	Default mode is 777
*/

ARRAY	*xcreat(larg,rarg)
ARRAY	*larg, *rarg;
{
	return mkintg(creat((char*)getfilename(rarg), larg?getint(larg):0777));
}

/*	l s e e k

	pointer <- lseek fildes offset whence
*/

ARRAY	*xlseek(larg,rarg)
ARRAY	*larg, *rarg;
{
	long		(*makelong)(), (*initlong())();
	int		fildes, whence;
	long		offset, lseek();
	unsigned	*base;
	ARRAY		*rslt, *mkarray();

	if(larg) error(SYNTAX);
	if(rarg->rank != 1) error(RANK);
	if(rarg->shape[0] != 3) error(LENGTH);
	
	base=rarg->shape+1;
	makelong= initlong(rarg);
	fildes=(int)(*makelong)(base, 0);
	offset=     (*makelong)(base, 1);
	whence=(int)(*makelong)(base, 2);

	offset = lseek(fildes, offset, whence);
	rslt= mkarray(LONG,0,0);
	*((long*)(rslt->shape)) = offset;
	freearray(rslt);
	return rslt;
}

/*	o p e n

	fildes <- {mode} open name	Default mode is read-only
*/

ARRAY	*xopen(larg,rarg)
ARRAY	*larg, *rarg;
{
	return mkintg(open((char*)getfilename(rarg), larg?getint(larg):0));
}

/*	r e a d

	buffer <- fildes read nbytes
*/

ARRAY	*xread(larg,rarg)
ARRAY	*larg, *rarg;
{
	int	nwant, ngot;
	ARRAY	*rslt, *mkarray();

	if(!larg) error(SYNTAX);
	nwant=getint(rarg);
	rslt=mkarray(CHAR,1,&nwant);	/* get buffer as char vector	*/
	ngot=read(getint(larg), (char*)(1+rslt->shape), nwant);
	freearray(rslt);		/* release buffer - still usable */
	if(ngot!=nwant)
		{if(ngot == -1) return mkintg(-1);
		rslt->shape[0]=ngot;	/* chars actually read	*/
		rslt->length = 3+ ((ngot+sizeof(int)-1)/sizeof(int));
		}
	return rslt;
}

/*	w r i t e

	nbytes <- fildes write buffer
*/

ARRAY	*xwrite(larg,rarg)
ARRAY	*larg, *rarg;
{
	if(!larg) error(SYNTAX);
	if(rarg->eltype != CHAR) error(DOMAIN);
	if(rarg->rank != 1) error(RANK);
	return mkintg(
		write(getint(larg), (char *)(1+rarg->shape), rarg->shape[0])
		);
}

/*	t r a n s l a t e

	translate text object
*/

ARRAY	*translate(larg,rarg,direction)
ARRAY	*larg, *rarg;
unsigned char	(*direction)();
{
	unsigned char	*cp, avtounix();
	unsigned	d, bound;

	if(larg) error(SYNTAX);
	if(rarg->eltype != CHAR) error(DOMAIN);

	for(bound=1, d=0; d < rarg->rank; d++) bound *= rarg->shape[d];
	cp=(unsigned char *)(rarg->shape+rarg->rank);
	while(bound--)
		{*cp = (*direction)(*cp);
		cp++;
		}
	return rarg;
}

/*	a v t o u n i x						*/

ARRAY	*xavtounix(larg,rarg)
ARRAY	*larg, *rarg;
{
	unsigned char	avtounix();

	return translate(larg,rarg,avtounix);
}

/*	u n i x t o a v						*/

ARRAY	*xunixtoav(larg,rarg)
ARRAY	*larg, *rarg;
{
	unsigned char	unixtoav();

	return translate(larg,rarg,unixtoav);
}

/*	Copyright (c) 1982 Dyadic Systems Limited	*/


