
/*	Copyright (c) 1982 Dyadic Systems Limited	*/

#include "apl.h"
#include "support.h"

#define	VERSION	"2.2"

static int	ppid;		/* parent id				*/

extern ARRAY	*avx(),
		*hex(),
		*ltom(),
		*ltov(),
		*mtol(),
		*dbr(),
		*ss(),
		*vtol();

static struct FUNCTION	
	{char		*f_name;
	unsigned	f_syntax;
	ARRAY		*(*f_pointer)();
	} functions[] =
		{
		"avx",		RSLT|MONA,	avx,
		"hex",		RSLT|MONA,	hex,
		"ltom",		RSLT|OPTL,	ltom,
		"ltov",		RSLT|OPTL,	ltov,
		"mtol",		RSLT|OPTL,	mtol,
		"dbr",		RSLT|MONA,	dbr,
		"ss",		RSLT|DYAD,	ss,
		"vtol",		RSLT|OPTL,	vtol,
		};

#define		NFUNCTIONS  (sizeof(functions)/sizeof(struct FUNCTION))

main(argc, argv)
int	argc;
char	*argv[];
{
	ARRAY		*getarray(), *freearray(), *(*f)();
	ARRAY		*larg, *rarg, *rslt;
	unsigned	s, i, fromapl();

	initxutils(argc, argv);
	larg = rarg = rslt = 0;

	setjmp(errbuff);		/* come here after error */
	for(;;)
		{larg = freearray(larg);
		rarg = freearray(rarg);
		rslt = freearray(rslt);

		if((i=fromapl()) >= NFUNCTIONS) syserr("main");
		s = functions[i].f_syntax;
		f = functions[i].f_pointer;
		switch(s & DYAD)
			{case DYAD:	/* dyadic and optional	*/
					larg = getarray();
					rarg = getarray();
					rslt = (*f)(larg,rarg);
					break;
			case MONA:	/* monadic */
					rarg = getarray();
					rslt = (*f)(rarg);
					break;
			default:	/* niladic */
					rslt = (*f)();
			}

		if(s & RSLT) putarray(rslt);
		}
}

/*	s y s e r r
	print syserr with (optional) message and die
*/
syserr(msg)
char	*msg;
{
	printf("xutils sys err%s %s\n", *msg ? ": " : "", msg);
	exit(1);
}

/*	i n i t x u t i l s
	define named functions or ALL if none named
*/
static	initxutils(argc, argv)
int	argc;
char	*argv[];
{
	register int	i;
	register char	*arg;
	int		nfns;
	char		wanted[NFUNCTIONS];

	/* preset functions requested flags	*/
	for(i=0; i<NFUNCTIONS; i++) wanted[i]=0;

	/* count and identify each function named in arguments	*/
	nfns = 0;
	while(argc--)
		{arg = *(argv++);
		for(i=0; i<NFUNCTIONS; i++)
			{if(wanted[i]) continue;
			if(!strcmp(arg,functions[i].f_name)) break;
			}
		if(i<NFUNCTIONS)
			{wanted[i] = 1;
			nfns++;
			}
		else if(strcmp(arg,"xutils") && strcmp(arg,""))
			printf("xutils %s not found %s\n",
				VERSION, arg);
		}
		
	/* define requested functions, or ALL if none requested	*/
	startup(nfns?nfns:NFUNCTIONS);
	for(i=0; i<NFUNCTIONS; i++)
		if(nfns==0 || wanted[i])
			define(functions[i].f_name, functions[i].f_syntax, i);
}

/*	a r r a y b o u n d
	number of elmts in an array
*/
unsigned	arraybound(array)
register ARRAY	*array;
{
	register unsigned	r, n;

	switch(array->rank)
		{case 0:	return 1;
		case 1:		return array->shape[0];
		default:	for(r=1, n=0; n<array->rank; n++)
					r *= array->shape[n];
				return r;
		}
}

/*	a r r a y s t a r t
	start of data part of array
*/
char	*arraystart(array)
ARRAY	*array;
{
	return (char*)(array->rank+array->shape);
}

/*	b l o c k m o v e
	character block move
*/
blockmove(from, to, nchars)
register char		*from, *to;
register unsigned	nchars;
{
	while(nchars--) *to++ = *from++;
}

/*	a v x
	return 0-origin index in quad av of simple text
*/

ARRAY	*avx(rarg)
ARRAY	*rarg;
{
	register char		*cp, *cpend;
	register unsigned	*ip;
	char			*cpstart;
	ARRAY			*rslt;

	if(rarg->eltype != CHAR) error(DOMAIN);
	cpstart = arraystart(rarg);
	cpend = cpstart + arraybound(rarg);

	/* if all characters are 127 or less can return SINT	*/

	for(cp=cpstart; cp<cpend; cp++) if( *cp & 128 ) break;
	if(cp==cpend)
		{rarg->eltype = SINT;
		return rarg;
		}
	rslt = mkarray(DINT,rarg->rank,rarg->shape);

	ip = (unsigned *) arraystart(rslt);
	for(cp=cpstart; cp<cpend; cp++) *ip++ = *cp & 0377;
	return rslt;
}

/*	h e x
	display object in hex
*/

static char	hexdigs[] = "0123456789ABCDEF";
static char	*hexdest;
static unsigned	counter;

#define	NL	0x3
#define	SP	0x4

ARRAY *hex(rarg)
ARRAY *rarg;
{
	ARRAY		*rslt;
	unsigned	shape;
	int		hexrslt(), hexcount(); 

	counter = 0;
	traverse(rarg,hexcount,0);
	shape = counter - 1;		/* dont need trailing delimiter */
	rslt = mkarray(CHAR,1,&shape);
	hexdest = (char*)(rslt->shape + 1);
	traverse(rarg,hexrslt,0);
	return rslt;
}

hexcount(array,depth)
ARRAY *array;
unsigned depth;
{
	counter += array->length * (
		2			/* two hex digits per byte   */
		* sizeof(int)		/* number of bytes per word  */
		+ 1			/* delimiter after each word */	
	);
}

hexrslt(array,depth)
ARRAY *array;
unsigned depth;
{
	register int bound, indx;
	register char *hexsrce;
	register unsigned char byte;

	bound = array->length * sizeof(int);	/* number of chars in array */
	hexsrce = (char*)array;			/* first word of array      */
	if (depth) *hexdest++ = NL;		/* newline before sub-array */
	for (indx = 0; indx<bound; indx++){
		if (
			indx &&			/* ... its not the first one */
			0 == indx%sizeof(int)	/* and its a word boundary   */
		) *hexdest++ = SP;		/* 	leave a gap	     */
		byte = *hexsrce++;		/* get target byte	     */
		*hexdest++ = hexdigs[byte>>4];	/* format first hex digit    */
		*hexdest++ = hexdigs[byte&0xf];	/* format second hex digit   */
	}
}

traverse(array,func,depth)  /* tramp (possibly nested) array and apply func */
ARRAY *array;
int (*func)();
unsigned depth;	/* depth of recursion */
{
	register unsigned indx, bound;

	(*func)(array,depth);
	switch (array->type){
	case SIMPLE:
		break;

	case NESTED:
		bound = array->length - 2;
		for (indx = array->rank; indx < bound ; indx++)
			traverse(array->shape[indx],func,depth+1);
		break;
	default:
		error(NONCE);
	}
}

/*	d b r
	remove leading, trailing, and multiple blanks
*/
ARRAY	*dbr(rarg)	
ARRAY	*rarg;
{
	register char	*iptr, *optr;
	register int	c, nelts, last, begun;
	char		*base;

	if(rarg->type != SIMPLE) error(DOMAIN);
	if(rarg->eltype != CHAR) error(DOMAIN);
	if(rarg->rank != 1) error(RANK);
	last = SP;
	begun = 0;
	base = optr = iptr = arraystart(rarg);
	for(nelts = rarg->shape[0]; nelts--;)
		{c = *iptr++;
		if(begun)
			{if(c != SP || last != SP)
				{*optr++ = last;
				last = c;
				}
			}
		else if(c != SP)
			{begun = 1;
			last = c;
			}
		}
	if(last != SP) *optr++ = last;

	nelts = rarg->shape[0] = optr-base;
	rarg->length = (nelts+sizeof(ARRAY)+sizeof(int)-1)/sizeof(int);
	return rarg;
}

/*	s s
	return boolean indicating where in left text arg occurs right txt arg
*/

ARRAY	*ss(larg,rarg)
ARRAY	*larg, *rarg;
{
	register char	*lptr, first, *lend, *lcp, *rcp, *rend;
	char		*lbase, *rbase;
	unsigned	lbd, rbd;

	if(larg->eltype != CHAR || rarg->eltype != CHAR) error(DOMAIN);

	lbd = arraybound(larg);
	rbd = arraybound(rarg);
	rbase = arraystart(rarg);
	rend = rbase + rbd;
	lptr = lbase = arraystart(larg);

	if(rbd && lbd >= rbd)
		/* loop through larg up to point of last possible match	*/
		for(first= *rbase, lend=lbase+lbd-rbd; lptr<=lend; lptr++)
			/* only enter nested loop if 1st char matches	*/
			if(first != *lptr) *lptr=0;
			else
				{for(lcp=lptr, rcp=rbase; ++lcp, ++rcp<rend;)
					if(*lcp != *rcp) break;
				*lptr = (rcp>=rend) ? 1 : 0;
				}
	/* fill in tail-end with zeros since too short to match	*/
	for(lend=lbase+lbd; lptr<lend; lptr++) *lptr = 0;

	/* result is larg itself, but numeric	*/
	larg->eltype = SINT;
	return larg;
}

/*	OLD VERSION RETURNED POSN OF FIRST MATCH AS SCALAR.....


ARRAY	*scalarint(i)
unsigned	i;
{
	ARRAY		*rslt, *mkarray();

	rslt = mkarray(DINT, 0, 0);
	rslt->shape[0] = i;
	return rslt;
}

ARRAY	*strsrch(larg,rarg)
ARRAY	*larg, *rarg;
{
	register char	*lptr, first, *lend, *lcp, *rcp, *rend;
	char		*lbase, *rbase;
	unsigned	lbound, rbound;

	if(larg->eltype != CHAR || rarg->eltype != CHAR) error(DOMAIN);
	if(larg->rank>1 || rarg->rank>1) error(RANK);

	lbase = (char*)(larg->shape + larg->rank);
	rbase = (char*)(rarg->shape + rarg->rank);
	lbound = larg->rank ? larg->shape[0] : 1;
	rbound = rarg->rank ? rarg->shape[0] : 1;
	rend = rbase + rbound;

	if(lbound>=rbound && rbound)
		{first = *rbase;
		lend = lbase + lbound - rbound;
		for(lptr=lbase; lptr<=lend; lptr++)
			{if(first != *lptr)continue;
			for(lcp=lptr, rcp=rbase; ++lcp, ++rcp<rend;)
				if (*lcp != *rcp ) break;
			if(rcp>=rend) return scalarint(1+lptr-lbase);
			}
		}
	return scalarint(1+lbound);
}
*/

/*	d e l i m i t e r
	return the character in an optional singleton array
*/
static char	delimiter(array)
ARRAY		*array;
{
	if(0==array)	return NL;
	if(array->eltype != CHAR) error(DOMAIN);
	if(1!=arraybound(array)) error(LENGTH);
	return *arraystart(array);
}

/*	m k v e c t o r
	return pointer to a text vector with specified text in it
*/
static ARRAY	*mkvector(from, upto)
register char	*from, *upto;
{
	ARRAY		*rslt;
	unsigned	n;

	if(upto<from) syserr("mkvector");
	n = upto-from;
	rslt = mkarray(CHAR, 1, &n);
	blockmove(from, arraystart(rslt), n);
	return rslt;
}

/*	u m a x
	unsigned max
*/
static unsigned	umax(a,b)
unsigned	a, b;
{
	return a>b ? a : b;
}

/*	l t o v
	linelist to vector of vectors
*/
ARRAY	*ltov(larg, rarg)
ARRAY	*larg, *rarg;
{
	register char	*cp, delim, *cpstart, *cpend;
	ARRAY		*rslt, **vptr;
	unsigned	ndelims;

	delim = delimiter(larg);
	if(rarg->eltype!=CHAR) error(DOMAIN);
	cpstart = arraystart(rarg);
	cpend = cpstart + arraybound(rarg);

	/* count delimiters	*/
	ndelims = 0;
	if(cpend != cpstart)
		{for(cp=cpstart; cp<cpend; cp++) if(delim == *cp) ndelims++;
		if(delim != *--cp) ndelims++; /* supply trailing delim	*/
		}
	rslt = mkarray(PNTR, 1, &ndelims);

	vptr = (ARRAY**) arraystart(rslt);
	if(ndelims)	/* install each vector in turn	*/
		{for(cp=cpstart; cp<cpend; cp++)
			{if(delim != *cp) continue;
			*vptr++ = mkvector(cpstart, cp);
			cpstart = 1+cp;
			}
		if(delim != *--cp)
			*vptr = mkvector(cpstart, ++cp);
		}
	else *vptr = mkvector(cpstart,cpstart);	/* prototype	*/
	return rslt;
}

/*	v t o l
	vector of vector to linelist
*/
ARRAY	*vtol(larg, rarg)
ARRAY	*larg, *rarg;
{
	register ARRAY	**vp, **vstart, **vend;
	ARRAY		*rslt;
	char		*cp, delim;
	unsigned	nchars;

	switch(rarg->eltype)
		{case CHAR:	return rarg;
		case PNTR:	break;
		default:	error(DOMAIN);
		}
	delim = delimiter(larg);
	vstart = (ARRAY**) arraystart(rarg);
	vend = vstart + arraybound(rarg);

	/* count characters & check single level nested text	*/
	nchars = vend-vstart;
	for(vp=vstart; vp<vend; vp++)
		{if(CHAR!= (*vp)->eltype) error(DOMAIN);
		nchars += arraybound(*vp);
		}
	rslt = mkarray(CHAR, 1, &nchars);

	/* copy in text, with trailing delimiter	*/
	cp = arraystart(rslt);
	for(vp=vstart; vp<vend; vp++)
		{nchars = arraybound(*vp);
		blockmove(arraystart(*vp), cp, nchars);
		cp+=nchars;
		*cp++ = delim;
		}

	return rslt;
}

/*	m t o l
	matrix to linelist with removal of trailing blanks
*/
ARRAY	*mtol(larg, rarg)
ARRAY	*larg, *rarg;
{
	register char	*cp, *to, *rowstart;
	char		delim, *rsltbase;
	ARRAY		*rslt;
	unsigned	ncols, nrows, n;

	delim = delimiter(larg);
	if(rarg->eltype!=CHAR) error(DOMAIN);
	nrows=ncols=1;
	if(rarg->rank)
		{for(n=0; n<rarg->rank-1; n++) nrows *= rarg->shape[n];
		ncols = rarg->shape[n];
		}
	n = nrows*(1+ncols);
	rslt = mkarray(CHAR, 1, &n);
	cp = rsltbase = arraystart(rslt);

	for(rowstart=arraystart(rarg); nrows--; rowstart+=ncols)
		{to = rowstart+ncols;
		while(--to >= rowstart && SP== *to);
		n = (1+to)-rowstart;
		blockmove(rowstart, cp, n);
		cp += n;
		*cp++ = delim;
		}
	rslt->shape[0] = n = cp - rsltbase;
	rslt->length = (n+sizeof(ARRAY)+sizeof(int)-1)/sizeof(int);
	return rslt;
}

/*	l t o m
	linelist to matrix
*/
ARRAY	*ltom(larg, rarg)
ARRAY	*larg, *rarg;
{
	register char	*cp, *rend, delim, *rstart, *last, *to;
	unsigned	shape[2], n;
	ARRAY		*rslt;

	delim = delimiter(larg);
	if(rarg->eltype != CHAR) error(DOMAIN);
	rstart = arraystart(rarg);
	rend = rstart + arraybound(rarg);

	/* count rows & longest line	*/
	shape[0]=shape[1]=0;
	for(last=cp=rstart; cp<rend; cp++)
		{if(delim != *cp) continue;
		shape[0]++;
		shape[1] = umax(shape[1], cp-last);
		last = 1+cp;
		}
	if(rstart!=rend && delim != *(cp-1))
		{shape[0]++;
		shape[1] = umax(shape[1], cp-last);
		}
	rslt = mkarray(CHAR, 2, shape);

	/* copy in rows, with trailing spaces	*/
	to = arraystart(rslt);
	for(last=cp=rstart; cp<rend; cp++)
		{if(delim != *cp) continue;
		n = cp-last;
		makerow(last,to,shape[1],n);
		to += shape[1];
		last = 1+cp;
		}
	if(rstart!=rend && delim != *(cp-1)) makerow(last,to,shape[1],cp-last);

	return rslt;
}

/*	m a k e r o w 
	copy n chars padding to length with spaces
*/
makerow(from, to, length, n)
register char		*from, *to;
register unsigned	length, n;
{
	length -= n;
	while(n--) *to++ = *from++;
	while(length--) *to++ = SP;
}

/*	Copyright (c) 1982 Dyadic Systems Limited	*/


